"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.prepareForInsertionContextless = exports.prepareArrayForInsertion = exports.prepareForInsertion = void 0;
const index_js_1 = require("../feature-libraries/index.js");
const toMapTree_js_1 = require("./toMapTree.js");
const internal_1 = require("@fluidframework/telemetry-utils/internal");
const index_js_2 = require("../util/index.js");
const index_js_3 = require("./core/index.js");
const internal_2 = require("@fluidframework/core-utils/internal");
/**
 * Prepare content from a user for insertion into a tree.
 * @remarks
 * This validates and converts the input, and if necessary invokes {@link prepareContentForHydration}.
 */
function prepareForInsertion(data, schema, destinationContext) {
    return prepareForInsertionContextless(data, schema, (0, index_js_1.getSchemaAndPolicy)(destinationContext), destinationContext.isHydrated() ? destinationContext : undefined);
}
exports.prepareForInsertion = prepareForInsertion;
/**
 * {@link prepareForInsertion} but batched for array content.
 * @privateRemarks
 * TODO:
 * Experimentally it was determined that making separate calls to prepareContentForHydration for each array item did not work.
 * This should be understood and fixed or have the factors that cause it clearly documented.
 * If fixed, this function should be removed, and arrays can just map over prepareForInsertion.
 */
function prepareArrayForInsertion(data, schema, destinationContext) {
    const mapTrees = data.map((item) => (0, toMapTree_js_1.mapTreeFromNodeData)(item, schema, destinationContext.isHydrated() ? destinationContext.nodeKeyManager : undefined, (0, index_js_1.getSchemaAndPolicy)(destinationContext)));
    if (destinationContext.isHydrated()) {
        prepareContentForHydration(mapTrees, destinationContext.checkout.forest);
    }
    return mapTrees;
}
exports.prepareArrayForInsertion = prepareArrayForInsertion;
/**
 * Split out from {@link prepareForInsertion} as to allow use without a context.
 * @remarks
 * Adding this entry point is a workaround for initialize not currently having a context.
 */
function prepareForInsertionContextless(data, schema, schemaAndPolicy, hydratedData) {
    const mapTree = (0, toMapTree_js_1.mapTreeFromNodeData)(data, schema, hydratedData?.nodeKeyManager, schemaAndPolicy);
    if (mapTree !== undefined && hydratedData !== undefined) {
        prepareContentForHydration([mapTree], hydratedData.checkout.forest);
    }
    return mapTree;
}
exports.prepareForInsertionContextless = prepareForInsertionContextless;
/**
 * A dummy key value used in {@link LocatedNodesBatch.rootPath} which will be replaced with the actual detached field once it is known.
 */
const placeholderKey = (0, index_js_2.brand)("placeholder");
/**
 * Records any proxies in the given content tree and does the necessary bookkeeping to ensure they are synchronized with subsequent reads of the tree.
 * @remarks If the content tree contains any proxies, this function must be called just prior to inserting the content into the tree.
 * Specifically, no other content may be inserted into the tree between the invocation of this function and the insertion of `content`.
 * The insertion of `content` must occur or else this function will cause memory leaks.
 * @param content - the content subsequence to be inserted, of which might deeply contain {@link TreeNode}s which need to be hydrated.
 * @param forest - the forest the content is being inserted into.
 * See {@link extractFactoryContent} for more details.
 */
function prepareContentForHydration(content, forest) {
    const batches = [];
    for (const item of content) {
        const batch = {
            rootPath: {
                parent: undefined,
                parentField: placeholderKey,
                parentIndex: 0,
            },
            paths: [],
        };
        batches.push(batch);
        walkMapTree(item, batch.rootPath, (p, node) => {
            batch.paths.push({ path: p, node });
        });
    }
    bindTreeNodes(batches, forest);
}
function walkMapTree(mapTree, path, onVisitTreeNode) {
    if ((0, index_js_3.tryUnhydratedFlexTreeNode)(mapTree)?.parentField.parent.parent !== undefined) {
        throw new internal_1.UsageError("Attempted to insert a node which is already under a parent. If this is desired, remove the node from its parent before inserting it elsewhere.");
    }
    const nexts = [];
    for (let next = [path, mapTree]; next !== undefined; next = nexts.pop()) {
        const [p, m] = next;
        const mapTreeNode = (0, index_js_3.tryUnhydratedFlexTreeNode)(m);
        if (mapTreeNode !== undefined) {
            const treeNode = index_js_3.unhydratedFlexTreeNodeToTreeNode.get(mapTreeNode);
            if (treeNode !== undefined) {
                onVisitTreeNode(p, treeNode);
            }
        }
        for (const [key, field] of m.fields) {
            for (const [i, child] of field.entries()) {
                nexts.push([
                    {
                        parent: p,
                        parentField: key,
                        parentIndex: i,
                    },
                    child,
                ]);
            }
        }
    }
}
/**
 * Register a collection of nodes with the forest so that they can be hydrated.
 * @param locatedNodes - the nodes to register with the forest.
 * Each index in this array expects its content to be added and produce its own `afterRootFieldCreated` event.
 * If array subsequence insertion is optimized to produce a single event, this will not work correctly as is, and will need to be modified to take in a single {@link LocatedNodesBatch}.
 */
function bindTreeNodes(locatedNodes, forest) {
    // Only subscribe to the event if there is at least one TreeNode tree to hydrate - this is not the case when inserting an empty array [].
    if (locatedNodes.length > 0) {
        // Creating a new array emits one event per element in the array, so listen to the event once for each element
        let i = 0;
        const off = forest.events.on("afterRootFieldCreated", (fieldKey) => {
            // Non null asserting here because of the length check above
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            const batch = locatedNodes[i];
            (0, internal_2.debugAssert)(() => batch.rootPath.parentField === placeholderKey);
            batch.rootPath.parentField = (0, index_js_2.brand)(fieldKey);
            for (const { path, node } of batch.paths) {
                (0, index_js_3.getKernel)(node).anchorProxy(forest.anchors, path);
            }
            if (++i === locatedNodes.length) {
                off();
            }
        });
    }
}
//# sourceMappingURL=prepareForInsertion.js.map